/**
 * Dns.c
 *
 * RFC 1035, 2821
 *
 * Copyright 2002, 2003 by Anthony Howe. All rights reserved.
 */

#undef NDEBUG

#ifndef DNS_DEFAULT_TIMEOUT
#define DNS_DEFAULT_TIMEOUT	5000
#endif

#ifndef DNS_DEFAULT_ATTEMPTS
#define DNS_DEFAULT_ATTEMPTS	4
#endif

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>

#if defined(__unix__) && ! defined(__CYGWIN__)
#  include <syslog.h>
#else
#  include <com/snert/lib/io/Log.h>
#endif

#include <com/snert/lib/io/Dns.h>
#include <com/snert/lib/io/posix.h>
#include <com/snert/lib/io/Socket.h>
#include <com/snert/lib/util/Text.h>

/***********************************************************************
 *** Constants
 ***********************************************************************/

#define DNS_PORT		53
#define UDP_PACKET_SIZE		512

#define TYPE_A			1
#define TYPE_NS			2
#define TYPE_CNAME		5
#define TYPE_SOA		6
#define TYPE_WKS		11	/* not supported */
#define TYPE_PTR		12
#define TYPE_HINFO		13	/* not supported */
#define TYPE_MINFO		14	/* not supported */
#define TYPE_MX			15
#define TYPE_TXT		16
#define TYPE_ALL		255

#define CLASS_IN		1
#define CLASS_CS		2
#define CLASS_CH		3
#define CLASS_HS		4

#define OP_QUERY		0x0000
#define OP_IQUERY		0x0800
#define OP_STATUS		0x1000

#define BITS_QR			0x8000	/* Query = 0, response = 1 */
#define BITS_OP			0x7800	/* op-code */
#define BITS_AA			0x0400	/* Response is authoritative. */
#define BITS_TC			0x0200	/* Message was truncated. */
#define BITS_RD			0x0100	/* Recursive query desired. */
#define BITS_RA			0x0080	/* Recursion available from server. */
#define BITS_Z			0x0070	/* Reserved - always zero. */
#define BITS_AU			0x0020	/* Answer authenticaed */
#define BITS_RCODE		0x000f	/* Response code */

#define SHIFT_QR		15
#define SHIFT_OP		11
#define SHIFT_AA		10
#define SHIFT_TC		9
#define SHIFT_RD		8
#define SHIFT_RA		7
#define SHIFT_Z			4
#define SHIFT_RCODE		0

#ifdef RCODE_ERRNO
#define RCODE_ERRNO		-1
#define RCODE_OK		0
#define RCODE_FORMAT		1
#define RCODE_SERVER		2
#define RCODE_UNDEFINED		3
#define RCODE_NOT_IMPLEMENTED	4
#define RCODE_REFUSED		5
#endif

#define DOMAIN_LENGTH		255

/***********************************************************************
 *** Internal types.
 ***********************************************************************/

struct header {
	unsigned short id;
	unsigned short bits;
	unsigned short qdcount;
	unsigned short ancount;
	unsigned short nscount;
	unsigned short arcount;
};

struct query {
	int length;
	struct header header;
	unsigned char data[UDP_PACKET_SIZE - sizeof (struct header)];
};

struct dns {
	int rcode;
	int attempts;
	long timeout;
	int nsockets;
	Socket *sockets;
	const char *error;
	struct query packet;
	unsigned short counter;
};

static long debug = 0;

/***********************************************************************
 *** Trivial accessors.
 ***********************************************************************/

/* These macros intended to retrieve network numeric data types stored
 * at odd memory addresses, which can cause some bus errors on certain
 * CPU types if the pointer is cast to a particular type.
 */
#define GET_NET_SHORT(v, p)	v = (unsigned short) (                     \
				    ((unsigned char *) p)[0] << CHAR_BIT   \
				  | ((unsigned char *) p)[1]               \
				)

#define GET_NET_LONG(v, p)	v = (unsigned long) (                      \
				    ((unsigned char *) p)[0] << CHAR_BIT*3 \
				  | ((unsigned char *) p)[1] << CHAR_BIT*2 \
				  | ((unsigned char *) p)[2] << CHAR_BIT   \
				  | ((unsigned char *) p)[3]               \
				)

#define GET_NET_SHORT_ADV(v, p)	v  = (unsigned short) (*((unsigned char *) p)++ << CHAR_BIT);   \
				v |= (unsigned short) (*((unsigned char *) p)++)

#define GET_NET_LONG_ADV(v, p)	v  = (unsigned long) (*((unsigned char *) p)++ << CHAR_BIT*3); \
				v |= (unsigned long) (*((unsigned char *) p)++ << CHAR_BIT*2); \
				v |= (unsigned long) (*((unsigned char *) p)++ << CHAR_BIT);   \
				v |= (unsigned long) (*((unsigned char *) p)++)

#define NET_SHORT_ADV(p)	(p) += 2

#define NET_LONG_ADV(p)		(p) += 4


void
DnsSetDebugMask(int mask)
{
	debug = mask;
}

const char *
DnsGetError(Dns dns)
{
	if (dns == (Dns) 0)
		return "null pointer to DNS structure";

#ifndef NDEBUG
	if (debug & DNS_DEBUG_GET)
		syslog(LOG_DEBUG, "DnsGetError(%lx) rcode=%d error=\"%s\"", (long) dns, ((struct dns *) dns)->rcode, ((struct dns *) dns)->error);
#endif

	return ((struct dns *) dns)->error;
}

int
DnsGetReturnCode(Dns dns)
{
	if (dns == (Dns) 0) {
		errno = EFAULT;
		return RCODE_NULL;
	}

#ifndef NDEBUG
	if (debug & DNS_DEBUG_GET)
		syslog(LOG_DEBUG, "DnsGetReturnCode(%lx) rcode=%d error=\"%s\"", (long) dns, ((struct dns *) dns)->rcode, ((struct dns *) dns)->error);
#endif

	return ((struct dns *) dns)->rcode;
}

void
DnsSetAttempts(Dns dns, int attempts)
{
	if (attempts < 1)
		attempts = 1;
#ifndef NDEBUG
	if (debug & DNS_DEBUG_SET)
		syslog(LOG_DEBUG, "DnsSetAttempts(%lx, %d)", (long) dns, attempts);
#endif
	((struct dns *) dns)->attempts = attempts * ((struct dns *) dns)->nsockets;
}

static void
DnsSetError(Dns dns, int rcode, const char *error)
{
#ifndef NDEBUG
	if (debug & DNS_DEBUG_SET)
		syslog(LOG_DEBUG, "DnsSetError(%lx, %d, '%s')", (long) dns, rcode, error);
#endif
	((struct dns *) dns)->rcode = rcode;
	((struct dns *) dns)->error = error;
}

void
DnsSetTimeout(Dns dns, long ms)
{
	if (ms < 1000)
		ms = 1000;

#ifndef NDEBUG
	if (debug & DNS_DEBUG_SET)
		syslog(LOG_DEBUG, "DnsSetTimeout(%lx, %ld)", (long) dns, ms);
#endif
	((struct dns *) dns)->timeout = ms;
}

/***********************************************************************
 *** Query handling.
 ***********************************************************************/

static const char *
buildQuery(struct dns *dns, const char *name, int type, int opcode)
{
	struct query *q;
	int length, needRoot;
	unsigned char *s, *t, *label;

#ifndef NDEBUG
	if (debug & DNS_DEBUG_TRACE)
		syslog(
			LOG_DEBUG,
			"buildQuery(dns=%lx, name=\"%s\", type=%d, opcode=%d)",
			(long) dns, name, type, opcode
		);
#endif
	length = (int) strlen(name);
	q = &((struct dns *) dns)->packet;
	needRoot = 0 < length && name[length-1] != '.' && name[length] == '\0';

	if (DOMAIN_LENGTH < length)
		return "DNS name length exceeds 255 octets";

	/* The header length, length of the name with root segment,
	 * and space for the type and class fields.
	 */
	q->length = sizeof (struct header) + 1 + length + needRoot + 4;

	/* Fill in the header fields that are not zero. */
	q->header.id = htons(++dns->counter);

	/* ASSUMPTION: Our local DNS servers support recursion. */
	q->header.bits = htons((unsigned short)((opcode & BITS_OP) | BITS_RD));

/* Without recursion I would have to perform alot of queries: first query
 * local DNS, which may return only the NS, then query authorative NS, which
 * may or may not support recursion and theory more queries for other info.
 */
/*	q->header.bits = htons((unsigned short)((opcode & BITS_OP))); */

	q->header.qdcount = htons(1);
	q->header.ancount = 0;
	q->header.nscount = 0;
	q->header.arcount = 0;

	/* Copy labels into question. */
	label = q->data;
	t = label + 1;
	for (s = (unsigned char *) name; *s != '\0' ; ++s, ++t) {
		if (*s == '.') {
			/* A label cannot exceed 63 octets in length. */
			if (63 < t - label - 1)
				return "DNS name segment exceeds 63 octets";

			/* Set label length. */
			*label = (unsigned char)(t - label - 1);
			label = t;
		} else {
			*t = *s;
		}
	}

	/* A label cannot exceed 63 octets in length. */
	if (63 < t - label - 1)
		return "DNS name segment exceeds 63 octets";

	/* Set last label length. */
	*label = (unsigned char)(t - label - 1);
	label = t;

	if (needRoot) {
		/* Set root label length. */
		*label++ = 0;
	}

	/* Complete question, assume the class is the Internet. */
	*label++ = 0;
	*label++ = (unsigned char) type;
	*label++ = 0;
	*label = CLASS_IN;

	return (const char *) 0;
}

static int
sendQuery(struct dns *dns)
{
	int attempt;
	Socket *thisSocket;
	long length, timeout, actualTimeout;

#ifndef NDEBUG
	if (debug & DNS_DEBUG_TRACE)
		syslog(LOG_DEBUG, "sendQuery(dns=%lx)", (long) dns);
#endif

	timeout = dns->timeout;
	thisSocket = dns->sockets;

	/* Using the defaults and assuming a single DNS server, then the
	 * exponential backoff algorithm will result in a max. delay of
	 * 75 seconds (5+10+20+40) to find an answer before giving up.
	 *
	 * With more than one server, the timeout each round is divided
	 * by the number of servers available. So for example:
	 *
	 *	2 servers: 5+5+ 5+5+ 10+10+ 20+20 = 80 seconds
	 *	3 servers: 5+5+5+ 5+5+5+ 6+6+6+ 13+13+13 = 87 seconds
	 */
	for (attempt = 0; attempt < dns->attempts; ++attempt) {
#ifndef NDEBUG
		if (debug & DNS_DEBUG_TRACE)
			syslog(LOG_DEBUG, "sendQuery() attempt %d via %lx", attempt, (long) *thisSocket);
#endif
		length = SocketWrite(*thisSocket, &dns->packet.header, dns->packet.length);

#if defined(__FreeBSD__)
/* sendto() is not defined to return EINVAL under FreeBSD, but under Linux yes.
 * See known issue http://www.freebsd.org/cgi/query-pr.cgi?pr=26506
 */
		if (SocketGetError(*thisSocket) == EINVAL)
			continue;
#endif

		if (length != dns->packet.length) {
			DnsSetError(dns, RCODE_ERRNO, "socket write error to DNS");
			return -1;
		}

		/* The base timeout value is divided by the number of DNS
		 * servers avaliable in order to limit the actual amount
		 * of time spent looking for an answer. Maintain a mimimum
		 * timeout value.
		 */
		if ((actualTimeout = timeout/dns->nsockets) < DNS_DEFAULT_TIMEOUT)
			actualTimeout= DNS_DEFAULT_TIMEOUT;

		if (SocketWaitForInput(*thisSocket, actualTimeout) == 0) {
			/*** Consider checking rcode here for RCODE_SERVER,
			 *** in which case continue loop.
			 ***/
			break;
		}

		/* Cycle through list of DNS servers. */
		if (*++thisSocket == (Socket) 0) {
			/* Reset cycle to the first server. */
			thisSocket = dns->sockets;

			/* Double the server timeout value this cycle. */
			timeout += timeout;
		}
	}

	if (dns->attempts <= attempt) {
		DnsSetError(dns, RCODE_ERRNO, "no response from DNS server");
		return -1;
	}

	memset(&dns->packet.header, 0, UDP_PACKET_SIZE);
	dns->packet.length = (int) SocketRead(*thisSocket, &dns->packet.header, UDP_PACKET_SIZE);
	if (dns->packet.length < 0) {
		DnsSetError(dns, RCODE_ERRNO, "socket read error from DNS");
		return -1;
	}

	dns->packet.header.id = ntohs(dns->packet.header.id);
	dns->packet.header.bits = ntohs(dns->packet.header.bits);

	if (dns->packet.header.id != dns->counter) {
		DnsSetError(dns, RCODE_FORMAT, "DNS request & response ID mismatch");
		return -1;
	}

	/* TODO: check if the message was truncated and redo the request
	 * over TCP as recommend by RFC 974.
	 */

	switch (dns->packet.header.bits & BITS_RCODE) {
	case RCODE_OK:
		DnsSetError(dns, RCODE_OK, "");
		break;
	case RCODE_FORMAT:
		DnsSetError(dns, RCODE_FORMAT, "DNS request format error");
		return -1;
	case RCODE_SERVER:
		DnsSetError(dns, RCODE_SERVER, "DNS server failure");
		return -1;
	case RCODE_UNDEFINED:
		DnsSetError(dns, RCODE_UNDEFINED, "DNS name not found");
		return -1;
	case RCODE_NOT_IMPLEMENTED:
		DnsSetError(dns, RCODE_NOT_IMPLEMENTED, "DNS operation not implemented");
		return -1;
	case RCODE_REFUSED:
		DnsSetError(dns, RCODE_REFUSED, "DNS request refused");
		return -1;
	default:
		DnsSetError(dns, RCODE_SERVER, "DNS server reported an unknown error");
		return -1;
	}

	dns->packet.header.qdcount = ntohs(dns->packet.header.qdcount);
	dns->packet.header.ancount = ntohs(dns->packet.header.ancount);
	dns->packet.header.nscount = ntohs(dns->packet.header.nscount);
	dns->packet.header.arcount = ntohs(dns->packet.header.arcount);

#ifdef IGNORE_OTHER_PARTS_OF_RESPONSE
	if (dns->packet.header.ancount <= 0) {
		DnsSetError(dns, RCODE_ERRNO, "no answers");
		return -1;
	}
#endif

	return 0;
}

/***********************************************************************
 *** Response Record Name Routines
 ***********************************************************************/

/*
 * Copy the string of the resource record name refered to by ptr
 * within the message into a buffer of length bytes. The name
 * labels are separated by dots and the final root label is added.
 * The terminating null character is appended.
 */
static void
copyName(unsigned char *message, unsigned char *ptr, unsigned char *buf, long length)
{
	for ( ; *ptr != 0; ptr += *ptr + 1) {
		if ((*ptr & 0xc0) == 0xc0) {
			unsigned short offset;
			GET_NET_SHORT(offset, ptr);
			ptr = &message[offset & 0x3fff];
		}

		/* Do we still have room in the buffer for the next label. */
		if (length <= *ptr)
			break;

		/* Copy label and advance pointers. */
		strncpy(buf, ptr+1, *ptr);
		length -= *ptr;
		buf += *ptr;
		*buf++ = '.';
	}

	if (0 < length)
		*buf = '\0';
}

/*
 * Return the string length of the resource record name refered to
 * by ptr within the message. The length includes for name label
 * (dot) delimiters and the final root label.
 */
static long
lengthName(unsigned char *message, unsigned char *ptr)
{
	long length;

	length = 0;
	for ( ; *ptr != 0; ptr += *ptr + 1) {
		if ((*ptr & 0xc0) == 0xc0) {
			unsigned long offset;
			GET_NET_SHORT(offset, ptr);
			ptr = &message[offset & 0x3fff];
		}

		length += *ptr + 1;
	}

	return length;
}

/*
 * Return a pointer to the next field within a resource record that
 * follows after the name field refered to by ptr.
 */
static unsigned char *
skipName(unsigned char *ptr)
{
	for ( ; ; ptr += *ptr + 1) {
		if (*ptr == 0) {
			++ptr;
			break;
		}
		if ((*ptr & 0xc0) == 0xc0) {
			ptr += 2;
			break;
		}
	}

	return ptr;
}

/*
 * Return an allocated C string for the resource record name refered
 * to by the ptr within the message. The pointer to the next unread
 * byte following the name field is passed back..
 */
static unsigned char *
getName(unsigned char *message, unsigned char *ptr, unsigned char **stop)
{
	long length;
	unsigned char *buf;

	length = lengthName(message, ptr);
	if (DOMAIN_LENGTH < length)
		return (unsigned char *) 0;

	if ((buf = malloc(length + 1)) == (unsigned char *) 0)
		return (unsigned char *) 0;

	copyName(message, ptr, buf, length + 1);

	if (stop != (unsigned char **) 0)
		*stop = skipName(ptr);

	return buf;
}

#ifdef DONT_NEED
/*
 * Return a pointer to the next resource record following the one
 * refered to by ptr.
 */
static unsigned char *
skipRecord(unsigned char *ptr)
{
	unsigned short length;

	ptr = skipName(ptr);
	NET_SHORT_ADV(ptr);		/* type */
	NET_SHORT_ADV(ptr);		/* class */
	NET_LONG_ADV(ptr);		/* ttl */
	GET_NET_SHORT_ADV(length, ptr);
	ptr += length;

	return ptr;
}
#endif

/***********************************************************************
 *** Response parsing.
 ***********************************************************************/

static Vector queryAndGetHosts(struct dns *dns, const char *name, int type);

void
DnsRecordRemoveEntry(void *record, void *data)
{
	DnsRecord entry = (DnsRecord) record;

	if (entry != (DnsRecord) 0) {
		if (entry->value != (char *) 0)
			free(entry->value);
		free(entry->name);
		free(entry);
	}
}

/*
 * Return true is we successfully allocated and parsed a resource
 * records that we want. The DnsRecord passed should be freed by
 * DnsRecordRemoveEntry().
 */
static int
getRecord(struct query *q, unsigned char **ptr, DnsRecord *record)
{
	DnsRecord entry;
	unsigned short length;

	entry = calloc(1, sizeof (struct DnsRecord));
	if (entry == NULL)
		return 0;

	*record = entry;

	entry->name = getName((unsigned char *) &q->header, *ptr, ptr);

	GET_NET_SHORT_ADV(entry->type, *ptr);
	NET_SHORT_ADV(*ptr); /* class (assume IN) */
	GET_NET_LONG_ADV(entry->ttl, *ptr);
	GET_NET_SHORT_ADV(length, *ptr);

	switch (entry->type) {
	case TYPE_NS:
		free(entry->name);
		entry->name = getName((unsigned char *) &q->header, *ptr, ptr);
		break;
	case TYPE_CNAME:
		entry->value = getName((unsigned char *) &q->header, *ptr, ptr);
		break;
	case TYPE_MX:
		free(entry->name);
		GET_NET_SHORT_ADV(entry->preference, *ptr);
		entry->name = getName((unsigned char *) &q->header, *ptr, ptr);
		break;
	case TYPE_A:
		GET_NET_LONG_ADV(entry->ip.s_addr, *ptr);
		/*** Leave it in network format ***/
		entry->ip.s_addr = htonl(entry->ip.s_addr);
		break;
	default:
		/* Skip this record. */
#ifndef NDEBUG
		if (debug & DNS_DEBUG_TRACE)
			syslog(LOG_DEBUG, "getRecord() entry->name=%s entry->type=%d skipped", entry->name, entry->type);
#endif
		*ptr += length;
		free(entry->name);
		free(entry);
		return 0;
	}

#ifndef NDEBUG
		if (debug & DNS_DEBUG_TRACE)
			syslog(LOG_DEBUG, "getRecord() entry->name=%s entry->type=%d", entry->name, entry->type);
#endif
	return 1;
}

static Vector
getRecordList(struct dns *dns)
{
	int i, j;
	char *find;
	Vector list;
	struct query *q;
	unsigned char *ptr;
	unsigned short type;
	DnsRecord entry, extra;

#ifndef NDEBUG
	if (debug & DNS_DEBUG_TRACE)
		syslog(LOG_DEBUG, "enter getRecordList(%lx)", (long) dns);
#endif

	/* Allow room for multi-homed hosts. */
	if ((list = VectorCreate(10)) == (Vector) 0)
		goto error0;

	VectorOnRemove(list, DnsRecordRemoveEntry, (void **) 0);
	q = &((struct dns *) dns)->packet;

	/* Skip question section. */
	ptr = q->data;
	for (i = 0; i < q->header.qdcount; i++) {
		ptr = skipName(ptr);
		/* ASSUMPTION: I'm assuming I only asked one question,
		 * so I remember the type of that question in case I
		 * have to repeat it below.
		 */
		GET_NET_SHORT_ADV(type, ptr);
		NET_SHORT_ADV(ptr);
	}

	/* Add all the returned resource-records of interest to the list. */
	j = q->header.ancount + q->header.nscount + q->header.arcount;
	for (i = 0; i < j; i++) {
		if (getRecord(q, &ptr, &entry))
			VectorAdd(list, entry);
	}

	/* Use the extra information provided with the answer to
	 * complete the answers; ie. resolve any CNAME and A records
	 * given with the answer.
	 */
	for (i = 0; i < q->header.ancount; ++i) {
		entry = VectorGet(list, i);
		if (entry->type == TYPE_A) {
#ifndef NDEBUG
			if (debug & DNS_DEBUG_TRACE)
				syslog(LOG_DEBUG, "entry->name=%s entry->type=%d (A)", entry->name, entry->type);
#endif
			continue;
		}

		if (entry->type == TYPE_CNAME) {
#ifndef NDEBUG
			if (debug & DNS_DEBUG_TRACE)
				syslog(LOG_DEBUG, "entry->name=%s entry->type=%d (CNAME)", entry->name, entry->type);
#endif
			if (1 < q->header.ancount) {
				/* We have more than one answer, assume that
				 * the other answers are the recusive result
				 * provided by the upstream DNS server. So
				 * just remove the CNAME from the result.
				 */
				VectorRemove(list, i, 1);
				q->header.ancount--;
				i--;
				continue;
			}

			/* RFC 974 "Issuing a Query" paragraph 4 about handling
			 * a CNAME response. We have to repeat the question with
			 * the domain name provided by the CNAME. Annoying, but
			 * legal.
			 *
			 * ASSUMPTION: Applies to NS records as well.
			 */
			find = TextDup(entry->value);
			VectorDestroy(list);
			list = queryAndGetHosts(dns, find, type);
			free(find);

			goto done;
		}

		find = entry->name;
#ifndef NDEBUG
		if (debug & DNS_DEBUG_TRACE)
			syslog(LOG_DEBUG, "entry->name=%s entry->type=%d", entry->name, entry->type);
#endif

		/* Loop through the extra records. */
		for (j = q->header.ancount + q->header.nscount; j < VectorLength(list); j++) {
			extra = VectorGet(list, j);

			if (TextInsensitiveCompare(find, extra->name) == 0) {
				if (extra->type == TYPE_CNAME) {
#ifndef NDEBUG
					if (debug & DNS_DEBUG_TRACE)
						syslog(LOG_DEBUG, "found %s CNAME %s", find, extra->value);
#endif
					find = extra->value;
					continue;
				}
				if (extra->type == TYPE_A) {
#ifndef NDEBUG
					if (debug & DNS_DEBUG_TRACE)
						syslog(LOG_DEBUG, "found %s A", find);
#endif
					entry->ip = extra->ip;
					break;
				}
#ifndef NDEBUG
				if (debug & DNS_DEBUG_TRACE)
					syslog(LOG_DEBUG, "found %s type=%d", find, extra->type);
#endif
			}
		}
	}

	/* Discard the extra information provided. */
	VectorRemove(list, q->header.ancount, VectorLength(list) -  q->header.ancount);

done:
error0:
#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "exit getRecordList(%lx) list=%lx", (long) dns, (long) list);
#endif
	return list;
}

static Vector
queryAndGetHosts(struct dns *dns, const char *name, int type)
{
	const char *error;
	Vector list = (Vector) 0;

#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "enter queryAndGetHosts(%lx, %s, %d) list=%lx", (long) dns, name, type, (long) list);
#endif

	error = buildQuery(dns, name, type, OP_QUERY);
	if (error != (const char *) 0) {
		DnsSetError(dns, RCODE_FORMAT, error);
		goto error0;
	}

	if (sendQuery(dns))
		goto error0;

	list = getRecordList(dns);
	if (list == (Vector) 0) {
		DnsSetError(dns, RCODE_ERRNO, "out of memory for IP address structure");
		goto error0;
	}
error0:
#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "exit queryAndGetHosts(%lx, %s, %d) list=%lx", (long) dns, name, type, (long) list);
#endif
	return list;
}

/***********************************************************************
 *** Principal DNS routines.
 ***********************************************************************/

Dns
DnsOpen(const char **servers)
{
	int i;
	struct dns *dns;

#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "enter DnsOpen(%lx)", (long) servers);
#endif
	for (i = 0; servers[i] != (const char *) 0; i++)
		;

	dns = malloc(sizeof *dns + (i+1) * sizeof (Socket));
	if (dns == (struct dns *) 0)
		goto error0;

	dns->sockets = (Socket *)(dns + 1);
	dns->nsockets = i;

	for (i = 0; i < dns->nsockets; i++) {
		dns->sockets[i] = SocketOpenUdp(servers[i], DNS_PORT);
		if (dns->sockets[i] == (Socket) 0)
			goto error1;
	}

	dns->sockets[i] = (Socket) 0;
	dns->counter = (unsigned short) rand();
	DnsSetTimeout(dns, DNS_DEFAULT_TIMEOUT);
	DnsSetAttempts(dns, DNS_DEFAULT_ATTEMPTS);

#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "exit DnsOpen(%lx) dns=%lx", (long) servers, (long) dns);
#endif
	return (Dns) dns;
error1:
	DnsClose(dns);
error0:
#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "error exit DnsOpen(%lx) dns=NULL errno=%d, %s", (long) servers, errno, strerror(errno));
#endif
	return (Dns) 0;
}

void
DnsClose(Dns dns)
{
	int i;
	struct dns *d = (struct dns *) dns;

#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "enter DnsClose(%lx)", (long) dns);
#endif
	if (dns != (Dns) 0) {
		for (i = 0; i < d->nsockets; i++)
			SocketClose(d->sockets[i]);
		free(d);
	}
#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "exit DnsClose(%lx)", (long) dns);
#endif
}

Vector
DnsGetAddresses(Dns dns, const char *domain)
{
	/*** Ideally we should check the known hosts list here
	 *** first before going to the trouble of a DNS query.
	 ***/
	return queryAndGetHosts(dns, domain, TYPE_A);
}

static int
CompareMxPreferences(const void *a, const void *b)
{
	return (*(DnsRecord *)a)->preference - (*(DnsRecord *)b)->preference;
}

static Vector
DnsGetMatchingIP(Dns dns, Vector list)
{
	int i;
	Vector Alist;
	DnsRecord listEntry, Aentry;

#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "enter DnsGetMatchingIP(%lx, %lx)", (long) dns, (long) list);
#endif

	/* Make sure we have the IPs for all the list members. */
	for (i = VectorLength(list); 0 <= --i; ) {
		listEntry = VectorGet(list, i);
#ifndef NDEBUG
		if (debug & DNS_DEBUG_TRACE) {
			char addr[50];
			ConvertIpToString(&listEntry->ip, addr, sizeof addr);
			syslog(LOG_DEBUG, "index=%d name=%s ip=%s", i, listEntry->name, addr);
		}
#endif
		if (listEntry->ip.s_addr != 0)
			continue;

#ifdef ALLOW_ENTRY_WITH_IP_STRING
/*
 * On Sunday, June 8, 2003, at 02:12 AM, Anthony Howe wrote:
 *
 * > Can't recall if I asked this before...
 * >
 * > By my reading of RFC 1035, section 3.3.9. MX RDATA format, the EXCHANGE
 * > field returns a host by name, not by IP address. So a DNS entry like the
 * > following is incorrectly specified, yes?  If yes, should a resolver
 * > identify such cases and make allowances, or should they just ignore the
 * > entry, consider it broken, and return a failure?
 *
 * Well, there's the "be conservative in what you send, liberal in what you
 * accept" policy which would seem to indicate that the resolver should, if
 * it can, figure out what was intended.
 *
 * That said, though, it's definitely broken, and a certain amount of logic
 * is present in the "it's invalid, skip it and move on" mindset.
 *
 * D
 */
		if (!SocketAddressToIP(MXentry->name, &MXentry->ip))
			continue;
#endif
		Alist = DnsGetAddresses(dns, listEntry->name);
		if (Alist != (Vector) 0 && 0 < VectorLength(Alist)) {
			Aentry = VectorGet(Alist, 0);
			listEntry->ip = Aentry->ip;
		} else {
			/* A record lookup failed. Note that the
			 * Return code and error message will have
			 * been modified. Fix below.
			 */
			VectorRemove(list, i, 1);
		}

		VectorDestroy(Alist);
	}

#ifdef HMM
/* Which is better? An empty list or a null pointer? */
	if (VectorLength(list) <= 0) {
		VectorDestroy(list);
		return (Vector) 0;
	}
#endif

	/* If we have at least one complete result, thats OK. */
	DnsSetError(dns, RCODE_OK, "");

#ifndef NDEBUG
	if (debug & DNS_DEBUG_ENTER_EXIT)
		syslog(LOG_DEBUG, "exit DnsGetMatchingIP(%lx, %lx)", (long) dns, (long) list);
#endif
	return list;
}

Vector
DnsGetMailServers(Dns dns, const char *domain)
{
	Vector MXlist;

	/* Get list of MX host names. */
	MXlist = queryAndGetHosts(dns, domain, TYPE_MX);
	if (MXlist == (Vector) 0 || DnsGetReturnCode(dns) != RCODE_OK){
		VectorDestroy(MXlist);
		return (Vector) 0;
	}

	if (VectorLength(MXlist) <= 0) {
		/* RFC 974 section "Interpreting the List of MX RRs"
		 * paragraph 2.
		 */
		return DnsGetAddresses(dns, domain);
	}

	VectorSort(MXlist, CompareMxPreferences);

	return DnsGetMatchingIP(dns, MXlist);
}

Vector
DnsGetNameServers(Dns dns, const char *domain)
{
	Vector NSlist;

	/* Get list of MX host names. */
	NSlist = queryAndGetHosts(dns, domain, TYPE_NS);
	if (NSlist == (Vector) 0 || DnsGetReturnCode(dns) != RCODE_OK) {
		VectorDestroy(NSlist);
		return (Vector) 0;
	}

	return DnsGetMatchingIP(dns, NSlist);
}

