/*
 * Time.h
 *
 * Fine resolution clock.
 *
 * Copyright 2002, 2003 by Anthony Howe.  All rights reserved.
 */

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#define UNKNOWN_API		0
#define ANSI_API		1	/* time(), seconds 10^1 */
#define BSD_API			2	/* gettimeofday(), micro-seconds 10^-6 */
#define POSIX_API		3	/* clock_gettime(), nano-seconds 10^-9 */

#include <stdlib.h>
#include <sys/types.h>

#include <com/snert/lib/version.h>
#include <com/snert/lib/sys/Time.h>

#if ! defined(TIME_API) || TIME_API == unknown
#error "No fine resolution clock support defined."
#endif

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

Time *
TimeGetNow()
{
	Time *t;

	if ((t = (Time *) malloc(sizeof *t)) != (Time *) 0)
		TimeSetNow(t);

	return t;
}

void
TimeSetNow(Time *t)
{
#if TIME_API == ANSI_API
	t->s = (unsigned long) time((time_t *) 0);
	t->ns = 0;
#endif
#if TIME_API == BSD_API
{
	struct timezone gmt = { 0, 0 };
	gettimeofday((struct timeval *) t, &gmt);
	t->ns *= UNIT_MILLI;
}
#endif
#if TIME_API == POSIX_API
	(void) clock_gettime(CLOCK_REALTIME, (struct timespec *) t);
#endif
}

Time *
TimeClone(Time *orig)
{
	Time *copy;

	if ((copy = (Time *) malloc(sizeof *copy)) != (Time *) 0)
		*copy = *orig;

	return copy;
}

void
TimeDestroy(Time *tp)
{
	if (tp != (Time *) 0)
		free(tp);
}

void
TimeAdd(Time *acc, Time *b)
{
	if (acc != (Time *) 0 && b != (Time *) 0) {
		acc->s += b->s;
		acc->ns += b->ns;

		if (UNIT_NANO <= acc->ns) {
			acc->ns -= UNIT_NANO;
			acc->s++;
		}
	}
}

void
TimeSub(Time *acc, Time *b)
{
	if (acc != (Time *) 0 && b != (Time *) 0) {
		if (acc->ns < b->ns) {
			acc->ns += UNIT_NANO;
			acc->s--;
		}
		acc->ns -= b->ns;
		acc->s -= b->s;
	}
}

int
TimeIsZero(Time *acc)
{
	return acc->s == 0 && acc->ns == 0;
}
