/*
 * mailout.c
 *
 * Copyright 2002, 2003 by Anthony Howe.  All rights reserved.
 */

#include <com/snert/lib/version.h>

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <com/snert/lib/io/Log.h>
#include <com/snert/lib/io/posix.h>
#include <com/snert/lib/io/Socket.h>
#include <com/snert/lib/util/EmailSpan.h>
#include <com/snert/lib/util/getopt.h>
#include <com/snert/lib/util/Text.h>

#define INPUT_LINE_SIZE		1000
#define HOST_NAME_SIZE		128
#define MAIL_ADDRESS_SIZE	256

int smtpPort = 25;
char *smtpHost = "127.0.0.1";

char usage[] =
"usage: mailout [-h host][-p port][-f[from]] to ... <message\n"
"\n"
"-h host\t\tSMTP host to contact [default localhost]\n"
"-p port\t\tSMTP port to use [default 25]\n"
"-f[from]\tEmail address of sender [default postmaster@localhost]\n"
"to ...\t\tOne or more destination email addresses\n"
"message\t\tThe message to send is read from standard input.\n"
"\n"
"mailout/1.9 Copyright 2002, 2003 by Anthony Howe.  All rights reserved.\n"
;

char line[INPUT_LINE_SIZE];
char hostname[HOST_NAME_SIZE];
char postmaster[MAIL_ADDRESS_SIZE];

void
printline(Socket s, char *line)
{
	int error;

	if (SocketPrint(s, line) < 0) {
		LogError("! SocketPrint() error");
		return;
	}

	/* Get and clear any errors. */
	if ((error = SocketGetError(s)) != 0) {
		LogError("! Socket error code: %d", error);
		return;
	}

	LogDebug("> %s", line);
}


int
getSmtpResponse(Socket s, int code, char *line, int linesize)
{
	int error;
	char *stop;
	long value;

	do {
		error = SocketWaitForInput(s, 120000);

		if (error == ETIMEDOUT) {
			LogError("! timeout before input from SMTP server");
			return 0;
		}

		if (error != 0) {
			LogError("! input error from SMTP server: %s (%d)", strerror(error), error);
			return 0;
		}

		value = SocketReadLine(s, line, linesize);

		if (value == 0) {
			LogWarn("! zero length read from SMTP server");
			return 0;
		}

		/* Did we read sufficient characters for a response code? */
		if (value < 4) {
			LogError("! read error from SMTP server: %s (%d)", strerror(errno), errno);
			return 0;
		}

		/* Get and clear any errors. */
		if ((error = SocketGetError(s)) != 0) {
			LogError("! socket error code: %s (%d)", strerror(error), error);
			return 0;
		}

		LogDebug("< %s", line);

		if ((value = strtol(line, &stop, 10)) == 421) {
			LogDebug("! SMTP server closing connection");
			return 0;
		}
	} while (line[3] == '-');

	return value == code && line + 3 == stop;
}

int
main(int argc, char **argv)
{
	int length;
	Socket client;
	char *from = (char *) 0;

#if defined(__BORLANDC__)
	_fmode = O_BINARY;
#endif

	LogSetLevel(LOG_DEBUG);
	LogSetProgramName("mailout");
	if (!LogOpen("mailout.log")) {
		fprintf(stderr, "Failed to open \"maillog.log\": %s (%d)\n", strerror(errno), errno);
		return 2;
	}

	while(getopt(argc, argv, "f;h:p:") != -1) {
		switch (optopt) {
		case 'f':
			from = optarg;
			if (from == (char *) 0)
				break;

			length = EmailSpanLocalPart(optarg);
			if (from[length] != '@') {
				LogError("-f argument \"%s\" not valid", from);
				LogError(usage);
				return 2;
			}

			optarg = &from[length + 1];
			length = EmailSpanDomainName(optarg, 1);
			if (optarg[length] != '\0') {
				LogError("-f argument \"%s\" not valid", from);
				LogError(usage);
				return 2;
			}
			break;
		case 'h':
			smtpHost = optarg;
			break;
		case 'p':
			smtpPort = strtol(optarg, (char **) 0, 10);
			break;
		default:
			fprintf(stderr, usage);
			fprintf(stderr, LIBSNERT_VERSION " " LIBSNERT_COPYRIGHT "\n");
			return 2;
		}
	}

	if (argc < optind + 1) {
		fprintf(stderr, usage);
		fprintf(stderr, LIBSNERT_VERSION " " LIBSNERT_COPYRIGHT "\n");
		return 2;
	}

	if (from == (char *) 0) {
		/* When no from address supplied, default to postmaster. */
		length = EmailSpanIPv4(smtpHost);
		sprintf(
			postmaster,
			smtpHost[length] == '\0'
				? "postmaster@[%s]"
				: "postmaster@%s",
			smtpHost
		);
		from = postmaster;
	}

	if (SocketInit()) {
		LogError("SocketInit()");
		return 1;
	}

	LogInfo("connecting to host=%s port=%d", smtpHost, smtpPort);

	client = SocketOpenTcpClient(smtpHost, smtpPort);
	if (client == (Socket) 0) {
		LogError("SocketOpenTcpClient()");
		return 1;
	}

/*	SocketSetTcpNoDelay(client, 1); */

	getSmtpResponse(client, 220, line, INPUT_LINE_SIZE);

	(void) gethostname(hostname, HOST_NAME_SIZE);
	sprintf(line, "HELO %s\r\n", hostname);
	printline(client, line);
	getSmtpResponse(client, 250, line, INPUT_LINE_SIZE);

	LogInfo("from=<%s>", from);
	sprintf(line, "MAIL FROM:<%s>\r\n", from);
	printline(client, line);
	getSmtpResponse(client, 250, line, INPUT_LINE_SIZE);

	for ( ; optind < argc; ++optind) {
		LogInfo("to=<%s>", argv[optind]);
		sprintf(line, "RCPT TO:<%s>\r\n", argv[optind]);
		printline(client, line);
		getSmtpResponse(client, 250, line, INPUT_LINE_SIZE);
	}

	printline(client, "DATA\r\n");
	getSmtpResponse(client, 354, line, INPUT_LINE_SIZE);

	while (0 <= (length = TextReadLine(FILENO_STDIN, line, INPUT_LINE_SIZE))) {
		if (line[0] == '.') {
			if (line[1] == '\0')
				break;
			SocketPrint(client, ".");
		}

		line[length++] = '\r';
		line[length++] = '\n';
		line[length] = '\0';

		printline(client, line);
	}

	printline(client, ".\r\n");
	getSmtpResponse(client, 250, line, INPUT_LINE_SIZE);

	printline(client, "QUIT\r\n");
#ifdef DONT_DO_THIS
	(void) SocketReadLine(client, line, INPUT_LINE_SIZE);
#endif
	SocketClose(client);

	LogDebug("done");

	return 0;
}


